"""oxtie package for storing data intelligently

The following are some useful places to look for more information:

  * oxtie.fronts: Package with implementations of front ends.
  * oxtie.backs:  Package with implementations of back ends.

"""

version = '0.1.1'
